from __future__ import print_function, unicode_literals
import fnmatch
import functools
import json
import os
import re
import sys

from nose.case import FunctionTestCase
from nose.plugins import Plugin
from nose.plugins.skip import SkipTest


def walk_subfolders(path):
    """
    Walk in the subtree within path and generate yield
    all subdirectories.
    """
    for dirpath, dirnames, filenames in os.walk(path):
        for dirname in dirnames:
            yield os.path.join(dirpath, dirname)


class SkipNose(Plugin):
    """
    Nose plugin class for skipnose

    Attributes
    ----------
    skipnose_include : list
        List of glob patterns to include
    skipnose_exclude : list
        List of glob patterns to exclude
    debug : bool
        Whether skipnose should print out debug messages
    """

    env_opt = 'NOSE_SKIPNOSE'
    env_include_opt = 'NOSE_SKIPNOSE_INCLUDE'
    env_exclude_opt = 'NOSE_SKIPNOSE_EXCLUDE'

    def __init__(self):
        super(SkipNose, self).__init__()
        self.debug = False
        self.skipnose_include = None
        self.skipnose_exclude = None
        self.skipnose_skip_tests = None

    def options(self, parser, env=os.environ):
        """
        Nose plugin method which allows to add shell options
        to nosetests runner
        """
        truth = ('true', '1', 'on')
        skipnose_on = env.get(self.env_opt, 'False').lower() in truth

        skip_include = list(filter(
            bool, re.split(r'[,;]', env.get(self.env_include_opt, ''))
        ))

        skip_exclude = list(filter(
            bool, re.split(r'[,;:]', env.get(self.env_exclude_opt, ''))
        ))

        parser.add_option(
            '--with-skipnose',
            action='store_true',
            default=skipnose_on,
            dest='skipnose',
            help='skipnose: enable skipnose nose plugin '
                 '(alternatively, set ${env}=1)'
                 ''.format(env=self.env_opt)
        )
        parser.add_option(
            '--skipnose-debug',
            action='store_true',
            default=False,
            dest='skipnose_debug',
            help='skipnose: enable debugging print-outs'
        )
        parser.add_option(
            '--skipnose-include',
            action='append',
            default=skip_include,
            dest='skipnose_include',
            help='skipnose: which directory to include in tests '
                 'using glob syntax.'
                 'Specifying multiple times will AND the clauses. '
                 'Single parameter ":" delimited clauses will be ORed. '
                 'Alternatively, set ${env} as [,;] delimited string for '
                 'AND and [:] for OR.'
                 ''.format(env=self.env_include_opt)
        )
        parser.add_option(
            '--skipnose-exclude',
            action='append',
            default=skip_exclude,
            dest='skipnose_exclude',
            help='skipnose: which directory to exclude in tests '
                 'using glob syntax.'
                 'Can be specified multiple times. '
                 '(alternatively, set ${env} as [,;:] delimited string)'
                 ''.format(env=self.env_exclude_opt)
        )
        parser.add_option(
            '--skipnose-skip-tests',
            action='store',
            dest='skipnose_skip_tests',
            help='skipnose: path to a json file which should contain '
                 'a list of test method names which should be skipped '
                 'under "skip_tests" key.'
        )

    def configure(self, options, conf):
        """
        When nosetests shell command is executed with correct
        parameters, nose calls configure method which allows
        to reflect the given flags in command
        in the plugin instance.

        One special boolean flag this method should set is
        ``enabled`` attribute which tells nosetests
        if this nose plugin should be enabled
        """
        if options.skipnose:
            self.enabled = True
            self.debug = options.skipnose_debug
            self.skipnose_include = list(map(
                lambda i: i.split(':'),
                options.skipnose_include
            ))
            self.skipnose_exclude = options.skipnose_exclude

            if options.skipnose_skip_tests:
                if not os.path.exists(options.skipnose_skip_tests):
                    print(
                        '{} not found'.format(options.skipnose_skip_tests),
                        file=sys.stderr
                    )
                    sys.exit(1)

                with open(options.skipnose_skip_tests, 'rb') as fid:
                    data = fid.read().decode('utf-8')
                    self.skipnose_skip_tests = json.loads(data)['skip_tests']

    def wantDirectory(self, dirname):
        """
        Nose plugin hook which allows to add logic whether nose
        should look for tests in the given directory.

        Even though nose allows the function to return both boolean
        ``True`` and ``False``, this implementation only returns
        ``False`` when a folder should explicitly be excluded and
        ``None`` when nose should use other factors to determine
        if the directory should be included. One example why
        ``True`` should not be returned is tests in folder
        without ``__init__.py`` file. If returned, all tests
        inside that folder will be executed even though is it
        not a valid Python package. By returning ``None``,
        that tells nose to apply additional logic in the
        determination such as checking if the folder is valid
        Python package.

        Parameters
        ----------
        dirname : str
            Directory path to consider

        Returns
        -------
        want : bool, None
            Boolean if tests should be executed inside the given folder.
            ``None`` is returned for unknown.
        """
        want = True
        basename = os.path.basename(dirname)

        if self.skipnose_include:
            want = all(map(
                lambda i: self._want_directory_by_includes(dirname, i),
                self.skipnose_include
            ))

        if self.skipnose_exclude and want is not False:
            # exclude the folder if the folder path
            # matches any of the exclude patterns
            want = not any(map(
                lambda i: fnmatch.fnmatch(basename, i),
                self.skipnose_exclude
            ))

        if self.debug:
            if not want:
                print('Skipnose: Skipping {}'.format(dirname), file=sys.stderr)
            else:
                print('Skipnose:          {}'.format(dirname), file=sys.stderr)

        # normalize boolean to only ``False`` or ``None``
        return False if want is False else None

    @staticmethod
    def _want_directory_by_includes(dirname, includes):
        # check all subfolders to see if any of them match
        # if yes, then this parent folder should be included
        # so that nose can get to the subfolder
        subfolders = map(
            os.path.basename,
            list(walk_subfolders(dirname)) + [dirname]
        )
        want = any(map(
            lambda i: fnmatch.filter(subfolders, i),
            includes
        ))

        # if directory is not wanted then there is a possibility
        # it is a subfolder of a wanted directory so
        # check against parent folder patterns
        if not want:
            parts = dirname.split(os.sep)
            want = any(map(
                lambda i: fnmatch.filter(parts, i),
                includes
            ))

        return want

    def startTest(self, test):
        """
        Skip tests when skipnose_skip_tests is provided
        """
        if not self.skipnose_skip_tests:
            return

        if isinstance(test.test, FunctionTestCase):
            test_name = '{}.{}'.format(
                test.test.test.__module__,
                test.test.test.__name__,
            )
        else:
            test_name = '{}.{}.{}'.format(
                test.test.__class__.__module__,
                test.test.__class__.__name__,
                test.test._testMethodName,
            )

        if test_name in self.skipnose_skip_tests:
            @functools.wraps(getattr(test.test, test.test._testMethodName))
            def skip_test(*args, **kwargs):
                raise SkipTest(
                    'Skipping {!r} as per --skipnose-skip-tests'
                    ''.format(test_name)
                )

            setattr(test.test, test.test._testMethodName, skip_test)



from __future__ import print_function, unicode_literals
from unittest import TestCase

import mock
from nose.case import FunctionTestCase
from nose.plugins.skip import SkipTest
from skipnose.skipnose import SkipNose, walk_subfolders


class TestWalkSubfolders(TestCase):
    @mock.patch('os.walk')
    def test_walk_subfolders(self, mock_walk):
        """
        Test that walk_subfolders returns os-joined paths
        as returned by os.walk
        """
        mock_walk.return_value = iter((
            ('path', ('a', 'b', 'c'), ('d', 'e', 'f')),
            ('other', ('q', 'w', 'e'), ('r', 't', 'y')),
        ))
        expected = [
            'path/a',
            'path/b',
            'path/c',
            'other/q',
            'other/w',
            'other/e',
        ]
        actual = walk_subfolders('foo')
        self.assertListEqual(list(actual), expected)
        mock_walk.assert_called_once_with('foo')


@mock.patch('skipnose.skipnose.print', mock.MagicMock(), create=True)
class TestSkipNoseConfig(TestCase):
    """
    Test class for skipnose configurations
    """

    def setUp(self):
        super(TestSkipNoseConfig, self).setUp()
        self.plugin = SkipNose()

    def test_options(self):
        """
        Test skipnose adds all configs to nose with
        correct defaults taken from environment variables.
        """
        env = {
            'NOSE_SKIPNOSE_INCLUDE': 'including',
            'NOSE_SKIPNOSE_EXCLUDE': 'excluding',
            'NOSE_SKIPNOSE': 'on',
        }
        mock_parser = mock.MagicMock()

        self.plugin.options(mock_parser, env)

        mock_parser.add_option.assert_has_calls(
            [
                mock.call('--with-skipnose',
                          action='store_true',
                          default=True,
                          dest=mock.ANY,
                          help=mock.ANY),
                mock.call('--skipnose-debug',
                          action='store_true',
                          default=False,
                          dest=mock.ANY,
                          help=mock.ANY),
                mock.call('--skipnose-include',
                          action='append',
                          default=['including'],
                          dest=mock.ANY,
                          help=mock.ANY),
                mock.call('--skipnose-exclude',
                          action='append',
                          default=['excluding'],
                          dest=mock.ANY,
                          help=mock.ANY),
            ]
        )

    @mock.patch('sys.exit')
    @mock.patch('os.path.exists')
    def test_configure(self, mock_path_exists, mock_sys_exit):
        """
        Test that configure sets class attributes correctly
        """
        mock_options = mock.MagicMock(
            skipnose_debug=mock.sentinel.debug,
            skipnose_include=['a', 'b:c'],
            skipnose_exclude=['x', 'y'],
            skipnose_skip_tests='foo.json',
        )
        mock_path_exists.return_value = True
        mock_open = mock.MagicMock()
        mock_open.return_value.__enter__.return_value.read.return_value = (
            b'{"skip_tests": ["one", "two"]}'
        )
        mock_sys_exit.side_effect = SystemExit

        with mock.patch('skipnose.skipnose.open', mock_open, create=True):
            self.plugin.configure(mock_options, None)

        self.assertTrue(self.plugin.enabled)
        self.assertEqual(self.plugin.debug, mock.sentinel.debug)
        self.assertEqual(self.plugin.skipnose_include, [['a'], ['b', 'c']])
        self.assertEqual(self.plugin.skipnose_exclude, ['x', 'y'])
        self.assertEqual(self.plugin.skipnose_skip_tests, ['one', 'two'])
        mock_open.assert_called_once_with('foo.json', 'rb')

    @mock.patch('sys.exit')
    @mock.patch('os.path.exists')
    def test_configure_error(self, mock_path_exists, mock_sys_exit):
        """
        Test that configure sets class attributes correctly when
        invalid skip-tests path is provided and nose exits
        """
        mock_options = mock.MagicMock(
            skipnose_debug=mock.sentinel.debug,
            skipnose_include=['a', 'b:c'],
            skipnose_exclude=['x', 'y'],
            skipnose_skip_tests='foo.data',
        )
        mock_path_exists.return_value = False
        mock_open = mock.MagicMock()
        mock_sys_exit.side_effect = SystemExit

        with mock.patch('skipnose.skipnose.open', mock_open, create=True):
            with self.assertRaises(SystemExit):
                self.plugin.configure(mock_options, None)

        self.assertTrue(self.plugin.enabled)
        self.assertEqual(self.plugin.debug, mock.sentinel.debug)
        self.assertEqual(self.plugin.skipnose_include, [['a'], ['b', 'c']])
        self.assertEqual(self.plugin.skipnose_exclude, ['x', 'y'])
        self.assertIsNone(self.plugin.skipnose_skip_tests)
        self.assertFalse(mock_open.called)
        mock_sys_exit.assert_called_once_with(1)


@mock.patch('skipnose.skipnose.print', mock.MagicMock(), create=True)
class TestSkipNose(TestCase):
    """
    Test class for skipnose functionality
    """

    def setUp(self):
        super(TestSkipNose, self).setUp()
        self.plugin = SkipNose()
        self.plugin.debug = True
        self.test_paths = [
            '/test',
            '/test/bar/cat/one',
            '/test/bar/cat/one/subone',
            '/test/bar/cat/two',
            '/test/bar/dog/one',
            '/test/bar/dog/one/api',
            '/test/bar/dog/one/api/subapi',
            '/test/bar/dog/one/api/subapi/moreapi',
            '/test/bar/dog/one/api/subapi/evenmoreapi',
            '/test/bar/dog/one/api/subapi/evenmoreapi/api',
            '/test/foo',
            '/test/foo/api',
            '/test/foo/api/subapi',
            '/test/foo/api/subapi/moreapi',
            '/test/foo/api/subapi/evenmoreapi',
            '/test/foo/api/subsubapi',
            '/test/foo/api/subsubapi/toomuchapi',
            '/test/foo/nonapi',
            '/test/foo/nonapi/folderone',
            '/test/foo/nonapi/foldertwo',
            '/test/foo/nonapi/foldertwo/morestuff',
            '/test/foo/nonapi/foldertwo/toomuch',
        ]

    def _mock_walk_subfolders(self, path):
        """
        Function to be provided to mock.side_effect to replace
        walk_subfolders functionality to use test paths.
        """
        paths = list(map(lambda i: i, self.test_paths))
        index = paths.index(path)
        if len(paths) > index + 1:
            return filter(lambda i: i.startswith(path), paths[index:])
        else:
            return []

    def _test_paths(self, valid):
        for path in self.test_paths:
            expected = None if path in valid else False
            actual = self.plugin.wantDirectory(path)
            self.assertEqual(actual, expected,
                             '{} != {} for {}'.format(actual, expected, path))

    @mock.patch('skipnose.skipnose.walk_subfolders')
    def test_want_directory_include(self, mock_walk_subfolders):
        """
        Test wantDirectory with include parameter
        """
        mock_walk_subfolders.side_effect = self._mock_walk_subfolders
        valid = [
            '/test',
            '/test/bar/dog/one',
            '/test/bar/dog/one/api',
            '/test/bar/dog/one/api/subapi',
            '/test/bar/dog/one/api/subapi/moreapi',
            '/test/bar/dog/one/api/subapi/evenmoreapi',
            '/test/bar/dog/one/api/subapi/evenmoreapi/api',
            '/test/foo',
            '/test/foo/api',
            '/test/foo/api/subapi',
            '/test/foo/api/subapi/moreapi',
            '/test/foo/api/subapi/evenmoreapi',
            '/test/foo/api/subsubapi',
            '/test/foo/api/subsubapi/toomuchapi',
        ]

        self.plugin.skipnose_include = [['api']]
        self._test_paths(valid)

    @mock.patch('skipnose.skipnose.walk_subfolders')
    def test_want_directory_include_multiple_or(self, mock_walk_subfolders):
        """
        Test wantDirectory with multiple include OR parameters
        """
        mock_walk_subfolders.side_effect = self._mock_walk_subfolders
        valid = [
            '/test',
            '/test/bar/dog/one',
            '/test/bar/dog/one/api',
            '/test/bar/dog/one/api/subapi',
            '/test/bar/dog/one/api/subapi/moreapi',
            '/test/bar/dog/one/api/subapi/evenmoreapi',
            '/test/bar/dog/one/api/subapi/evenmoreapi/api',
            '/test/foo',
            '/test/foo/api',
            '/test/foo/api/subapi',
            '/test/foo/api/subapi/moreapi',
            '/test/foo/api/subapi/evenmoreapi',
            '/test/foo/api/subsubapi',
            '/test/foo/api/subsubapi/toomuchapi',
            '/test/foo/nonapi',
            '/test/foo/nonapi/folderone',
            '/test/foo/nonapi/foldertwo',
            '/test/foo/nonapi/foldertwo/morestuff',
            '/test/foo/nonapi/foldertwo/toomuch',
        ]

        self.plugin.skipnose_include = [['api', 'foo']]
        self._test_paths(valid)

    @mock.patch('skipnose.skipnose.walk_subfolders')
    def test_want_directory_include_multiple_and(self, mock_walk_subfolders):
        """
        Test wantDirectory with multiple include OR parameters
        """
        mock_walk_subfolders.side_effect = self._mock_walk_subfolders
        valid = [
            '/test',
            '/test/foo',
            '/test/foo/api',
            '/test/foo/api/subapi',
            '/test/foo/api/subapi/moreapi',
            '/test/foo/api/subapi/evenmoreapi',
            '/test/foo/api/subsubapi',
            '/test/foo/api/subsubapi/toomuchapi',
        ]

        self.plugin.skipnose_include = [['api'], ['foo']]
        self._test_paths(valid)

    def test_want_directory_exclude(self):
        """
        Test wantDirectory with exclude parameter
        """
        valid = [
            '/test',
            '/test/bar/cat/one',
            '/test/bar/cat/one/subone',
            '/test/bar/cat/two',
            '/test/bar/dog/one',
            '/test/bar/dog/one/api/subapi',  # implicitly skipped by walk
            '/test/bar/dog/one/api/subapi/moreapi',  # noqa implicitly skipped by walk
            '/test/bar/dog/one/api/subapi/evenmoreapi',  # noqa implicitly skipped by walk
            '/test/foo',
            '/test/foo/api/subapi',  # implicitly skipped by walk
            '/test/foo/api/subapi/moreapi',  # implicitly skipped by walk
            '/test/foo/api/subapi/evenmoreapi',  # implicitly skipped by walk
            '/test/foo/api/subsubapi',  # implicitly skipped by walk
            '/test/foo/api/subsubapi/toomuchapi',  # implicitly skipped by walk
            '/test/foo/nonapi',
            '/test/foo/nonapi/folderone',
            '/test/foo/nonapi/foldertwo',
            '/test/foo/nonapi/foldertwo/morestuff',
            '/test/foo/nonapi/foldertwo/toomuch',
        ]
        self.plugin.skipnose_exclude = ['api']
        self._test_paths(valid)

    def test_want_directory_exclude_multiple(self):
        """
        Test wantDirectory with multiple exclude parameter
        """
        valid = [
            '/test',
            '/test/bar/cat/one',
            '/test/bar/cat/one/subone',
            '/test/bar/cat/two',
            '/test/bar/dog/one',
            '/test/bar/dog/one/api/subapi',  # implicitly skipped by walk
            '/test/bar/dog/one/api/subapi/moreapi',  # noqa implicitly skipped by walk
            '/test/bar/dog/one/api/subapi/evenmoreapi',  # noqa implicitly skipped by walk
            '/test/foo/api/subapi',  # implicitly skipped by walk
            '/test/foo/api/subapi/moreapi',  # implicitly skipped by walk
            '/test/foo/api/subapi/evenmoreapi',  # implicitly skipped by walk
            '/test/foo/api/subsubapi',  # implicitly skipped by walk
            '/test/foo/api/subsubapi/toomuchapi',  # implicitly skipped by walk
            '/test/foo/nonapi',  # implicitly skipped by walk
            '/test/foo/nonapi/folderone',  # implicitly skipped by walk
            '/test/foo/nonapi/foldertwo',  # implicitly skipped by walk
            '/test/foo/nonapi/foldertwo/morestuff',  # noqa implicitly skipped by walk
            '/test/foo/nonapi/foldertwo/toomuch',  # implicitly skipped by walk
        ]
        self.plugin.skipnose_exclude = ['api', 'foo']
        self._test_paths(valid)

    def test_start_test_no_tests_to_skip(self):
        self.plugin.skipnose_skip_tests = None

        self.assertIsNone(self.plugin.startTest(mock.Mock()))

    def test_start_test_function_test_case(self):
        self.plugin.skipnose_skip_tests = ['one', 'two', 'foo.bar']

        def test():
            """"""

        mock_test = mock.MagicMock()
        mock_test.test = mock.MagicMock(spec=FunctionTestCase)
        mock_test.test.test = mock.MagicMock(__module__='foo', __name__='bar')
        mock_test.test._testMethodName = 'method'
        mock_test.test.method = test

        self.plugin.startTest(mock_test)

        replaced_method = mock_test.test.method
        self.assertIsNot(replaced_method, test)
        self.assertTrue(callable(replaced_method))
        with self.assertRaises(SkipTest):
            replaced_method()

    def test_start_test_method_test_case(self):
        self.plugin.skipnose_skip_tests = ['one', 'two', 'foo.Foo.method']

        class Foo(object):
            def method(self):
                """"""

        Foo.__module__ = 'foo'
        instance = Foo()
        test = instance.method

        mock_test = mock.MagicMock()
        mock_test.test = instance
        mock_test.test._testMethodName = 'method'
        mock_test.test.method = test

        self.plugin.startTest(mock_test)

        replaced_method = mock_test.test.method
        self.assertIsNot(replaced_method, test)
        self.assertTrue(callable(replaced_method))
        with self.assertRaises(SkipTest):
            replaced_method()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function

from .skipnose import SkipNose  # noqa


__author__ = 'Miroslav Shubernetskiy'
__email__ = 'miroslav@miki725.com'
__version__ = '0.3.0'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function
import os
import re
from setuptools import setup, find_packages


def read(fname):
    return (open(os.path.join(os.path.dirname(__file__), fname), 'rb')
            .read().decode('utf-8'))


authors = read('AUTHORS.rst')
history = read('HISTORY.rst').replace('.. :changelog:', '')
licence = read('LICENSE.rst')
readme = read('README.rst')


module_file = read("skipnose/__init__.py")
metadata = dict(re.findall("__([a-z]+)__\s*=\s*'([^']+)'", module_file))


requirements = read('requirements.txt').splitlines() + [
    'setuptools',
]


test_requirements = (
    read('requirements.txt').splitlines() +
    read('requirements-dev.txt').splitlines()[1:]
)


setup(
    name='skipnose',
    version=metadata['version'],
    author=metadata['author'],
    description='Nose plugin which allows to include/exclude directories '
                'for testing by their glob pattern.',
    long_description='\n\n'.join([readme, history, authors, licence]),
    url='https://github.com/Dealertrack/skipnose',
    packages=find_packages(exclude=['test', 'test.*']),
    install_requires=requirements,
    test_suite='tests',
    tests_require=test_requirements,
    entry_points={
        'nose.plugins.0.10': [
            'NOSETESTS_SKIPNOSE = skipnose:SkipNose'
        ]
    },
    keywords=' '.join([
        'skipnose',
        'test',
        'nosetests',
        'nose',
        'nosetest',
    ]),
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Testing",
    ],
)
